/**
 * ImageJ Plugin using JCuda
 * 
 * Copyright (c) 2013 Marco Hutter - http://www.jcuda.org
 */

import static jcuda.runtime.JCuda.cudaFree;
import static jcuda.runtime.JCuda.cudaMalloc;
import static jcuda.runtime.JCuda.cudaMemcpy;
import static jcuda.runtime.cudaMemcpyKind.cudaMemcpyDeviceToHost;
import static jcuda.runtime.cudaMemcpyKind.cudaMemcpyHostToDevice;
import ij.IJ;
import ij.ImagePlus;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import jcuda.Pointer;
import jcuda.Sizeof;
import jcuda.utils.KernelLauncher;

/**
 * A simple example for an ImageJ Plugin that uses JCuda.
 */
public class Simple_JCuda_Plugin implements PlugInFilter
{
	/**
	 * The current image to operate on
	 */
    private ImagePlus currentImage = null;
    
    /**
     * The KernelLauncher which will be used to launch the
     * CUDA kernel that was read from a PTX file.
     */
    private KernelLauncher kernelLauncher = null;
    
    @Override
    public void run(ImageProcessor imageProcessor) 
    {
    	int[] pixels = (int[])imageProcessor.getPixels();
        int w = imageProcessor.getWidth();
        int h = imageProcessor.getHeight();
        execute(pixels, w, h);
        currentImage.updateAndDraw();
    }

    /**
     * Will execute the CUDA kernel with the given parameters
     * 
     * @param pixels An array containing the pixels of the
     * image as RGB integers
     * @param w The width of the image
     * @param h The height of the image
     */
    void execute(int pixels[], int w, int h)
    {
    	// Allocate memory on the device, and copy the host data to the device 
        int size = w*h*Sizeof.INT;
        Pointer pointer = new Pointer();
        cudaMalloc(pointer, size);
        cudaMemcpy(pointer, Pointer.to(pixels), size, cudaMemcpyHostToDevice);

        // Set up and call the kernel
        int blockSize = 16;
        int gridSize = (int)Math.ceil((double)Math.max(w, h)/blockSize);
        kernelLauncher.setGridSize(gridSize, gridSize);
        kernelLauncher.setBlockSize(blockSize, blockSize, 1);
        kernelLauncher.call(pointer, w, h);
        
        // Copy the data from the device back to the host and clean up
        cudaMemcpy(Pointer.to(pixels), pointer, size, cudaMemcpyDeviceToHost);
        cudaFree(pointer);
    }

    @Override
    public int setup(String arg, ImagePlus imagePlus)
    {
    	if (arg != null && arg.equals("about"))
    	{
            IJ.showMessage(
                    "About Simple JCuda Plugin...",
                    "An example of an ImageJ plugin using JCuda\n");
    		return DOES_RGB;
    	}
        currentImage = imagePlus;

        // Obtain the CUDA source code from the CUDA file
        String cuFileName = "simpleJCudaPluginKernel.cu";
        InputStream cuInputStream = 
        	getClass().getResourceAsStream(cuFileName);
        String sourceCode = "";;
        try
        {
            sourceCode = read(cuInputStream);
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
        finally
        {
            try
            {
                cuInputStream.close();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }
        
        // Create the kernelLauncher that will execute the kernel
    	kernelLauncher = KernelLauncher.compile(sourceCode, "invert");
    	
    	return DOES_RGB;
    }
    
    /**
     * Reads the contents for the given input stream and returns
     * it as a String
     * 
     * @param inputStream The input stream
     * @return The contents of the input stream
     * @throws IOException If an IO error occurs
     */
    private static String read(InputStream inputStream) throws IOException
    {
        BufferedReader br = new BufferedReader(
            new InputStreamReader(inputStream));
        StringBuilder sb = new StringBuilder();
        while (true)
        {
            String line = br.readLine();
            if (line == null)
            {
                break;
            }
            sb.append(line).append("\n");
        }
        return sb.toString();
    }

}